/* Common routines for ip4 addresses. */

#include "ip4addr.h"

#define digit(c) ((c) >= '0' && (c) <= '9')
#define d2n(c) ((c) - '0')

#define cret(v, np, s) ((np) ? *(np) = (char*)(s), (v) : *(s) ? 0 : (v))
#define eret(np, s) ((np) ? *(np) = (char*)(s), 0 : 0)

static unsigned ip4mask(const char *s, ip4addr_t a, char **np) {
  /* helper routine for ip4cidr() */
  unsigned b;
  if (!digit(*s))
    return eret(np, s);
  b = d2n(*s++);
  while(digit(*s))
    if ((b = b * 10 + d2n(*s++)) > 32)
      return eret(np, s - 1);
  if (!b || a & ((((ip4addr_t)1) << (32 - b)) - 1))
    return eret(np, s - 1);
  return cret(b, np, s);
}

/* parse ip4 CIDR range in `s', store base in *ap and
 * return number of bits */

unsigned ip4cidr(const char *s, ip4addr_t *ap, char **np) {
  ip4addr_t o;
  *ap = 0;

#define ip4oct(bits)					\
  if (!digit(*s))					\
    return eret(np, s);					\
  o = d2n(*s++);					\
  while(digit(*s))					\
    if ((o = o * 10 + d2n(*s++)) > 255)			\
      return eret(np, s - 1);				\
  *ap |= o << bits;					\
  if (*s == '/')					\
    return ip4mask(s + 1, *ap, np)

  ip4oct(24);
  if (*s++ != '.')
    return cret(8, np, s-1);
  ip4oct(16);
  if (*s++ != '.')
    return cret(16, np, s-1);
  ip4oct(8);
  if (*s++ != '.')
    return cret(24, np, s-1);
  ip4oct(0);
  return
    *s == '/' ? ip4mask(s+1, *ap, np) : cret(32, np, s);
}

#ifdef TEST
#include <stdio.h>

int main(int argc, char **argv) {
  int i;
  ip4addr_t a;
  unsigned b;
  char *np;
  for(i = 1; i < argc; ++i) {
    printf("%s:", argv[i]);
    b = ip4cidr(argv[i], &a, NULL);
    if (!b)
      printf(" err;");
    else
      printf(" bits=%u ip=%u.%u.%u.%u;",
             b, a>>24, (a>>16)&0xff, (a>>8)&0xff, a&0xff);
    b = ip4cidr(argv[i], &a, &np);
    if (!b)
      printf(" err tail=`%s'\n", np);
    else
      printf(" bits=%u ip=%u.%u.%u.%u tail=`%s'\n",
             b, a>>24, (a>>16)&0xff, (a>>8)&0xff, a&0xff, np);
   }
  return 0;
}

#endif
